#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "graph.h"
#include "fastme.h"


boolean leaf(node *v);
edge *siblingEdge(edge *e);
edge *depthFirstTraverse(tree *T, edge *e);
edge *findBottomLeft(edge *e);
edge *topFirstTraverse(tree *T, edge *e);
edge *moveUpRight(edge *e);
void WFext(edge *e, double **A);
void WFint(edge *e, double **A);

void limitedFillTableUp(edge *e, edge *f, double **A, edge *trigger);
void assignBalWeights(tree *T, double **A);
void updateAveragesMatrix(tree *T, double **A, node *v,int direction);
void bNNItopSwitch(tree *T, edge *e, int direction, double **A);
int bNNIEdgeTest(edge *e, tree *T, double **A, double *weight);
void updatePair(double **A, edge *nearEdge, edge *farEdge, node *v, node *root, double dcoeff, int direction);

int *initPerm(int size);

void reHeapElement(int *p, int *q, double *v, int length, int i);
void pushHeap(int *p, int *q, double *v, int length, int i);
void popHeap(int *p, int *q, double *v, int length, int i);


void bNNIRetestEdge(int *p, int *q, edge *e,tree *T, double **avgDistArray, 
		double *weights, int *location, int *possibleSwaps)
{
  int tloc;
  tloc = location[e->head->index+1];
  location[e->head->index+1] = 
    bNNIEdgeTest(e,T,avgDistArray,weights + e->head->index+1);
  if (NONE == location[e->head->index+1])
    {
      if (NONE != tloc)
	popHeap(p,q,weights,(*possibleSwaps)--,q[e->head->index+1]);	  
    }
  else
    {
      if (NONE == tloc)
	pushHeap(p,q,weights,(*possibleSwaps)++,q[e->head->index+1]);
      else
	reHeapElement(p,q,weights,*possibleSwaps,q[e->head->index+1]);
    }
}

int makeThreshHeap(int *p, int *q, double *v, int arraySize, double thresh);

void permInverse(int *p, int *q, int length);

void bNNI(tree *T, double **avgDistArray, int *count)
{
  edge *e, *centerEdge;
  edge **edgeArray;
  int *p, *location, *q;
  int i;
  int possibleSwaps;
  double *weights;
  p = initPerm(T->size+1);
  q = initPerm(T->size+1);
  edgeArray = (edge **) malloc((T->size+1)*sizeof(double));
  weights = (double *) malloc((T->size+1)*sizeof(double));
  location = (int *) malloc((T->size+1)*sizeof(int));
  for (i=0;i<T->size+1;i++)
    {
      weights[i] = 0.0;
      location[i] = NONE;
    }
  e = findBottomLeft(T->root->leftEdge); 
  while (NULL != e)
    {
      edgeArray[e->head->index+1] = e;
      location[e->head->index+1] = 
	bNNIEdgeTest(e,T,avgDistArray,weights + e->head->index + 1);
      e = depthFirstTraverse(T,e);
    } 
  possibleSwaps = makeThreshHeap(p,q,weights,T->size+1,0.0);
  permInverse(p,q,T->size+1);
  /*we put the negative values of weights into a heap, indexed by p
    with the minimum value pointed to by p[1]*/
  /*p[i] is index (in edgeArray) of edge with i-th position 
    in the heap, q[j] is the position of edge j in the heap */
  while (weights[p[1]] < 0)
    {
      centerEdge = edgeArray[p[1]];
      (*count)++;
      T->weight = T->weight + weights[p[1]];
      bNNItopSwitch(T,edgeArray[p[1]],location[p[1]],avgDistArray);
      location[p[1]] = NONE;
      weights[p[1]] = 0.0;  /*after the bNNI, this edge is in optimal
			      configuration*/
      popHeap(p,q,weights,possibleSwaps--,1);
      /*but we must retest the other four edges*/
      e = centerEdge->head->leftEdge;
      bNNIRetestEdge(p,q,e,T,avgDistArray,weights,location,&possibleSwaps);
      e = centerEdge->head->rightEdge;
      bNNIRetestEdge(p,q,e,T,avgDistArray,weights,location,&possibleSwaps);
      e = siblingEdge(centerEdge);
      bNNIRetestEdge(p,q,e,T,avgDistArray,weights,location,&possibleSwaps);
      e = centerEdge->tail->parentEdge;
      bNNIRetestEdge(p,q,e,T,avgDistArray,weights,location,&possibleSwaps);
    }
  free(p);
  free(q);
  free(location);
  free(edgeArray);
  assignBalWeights(T,avgDistArray);
}




void updateSubTreeAfterNNI(double **A, node *v, edge *nearEdge, node *r1, node *r2,
			   double dcoeff, int direction)
{
  edge *sib;
  switch(direction)
    {
    case UP: /*nearEdge is below the center edge of the NNI*/
      if (NULL != nearEdge->head->leftEdge)
	updateSubTreeAfterNNI(A, v, nearEdge->head->leftEdge, r1, r2, 0.5*dcoeff,UP);
      if (NULL != nearEdge->head->rightEdge)
	updateSubTreeAfterNNI(A, v, nearEdge->head->rightEdge, r1, r2, 0.5*dcoeff,UP);
      updatePair(A, nearEdge, nearEdge, r1, r2, dcoeff, UP);
      A[nearEdge->head->index][v->index] = 
	A[v->index][nearEdge->head->index] = 
	A[v->index][nearEdge->head->index] + 
	0.5*A[nearEdge->head->index][r2->index] -
	0.5*A[nearEdge->head->index][r1->index];    
      break;
    case DOWN: /*nearEdge is above the center edge of the NNI*/
      sib = siblingEdge(nearEdge);
      if (NULL != sib)
	updateSubTreeAfterNNI(A, v, sib, r1, r2, 0.5*dcoeff, SKEW);
      if (NULL != nearEdge->tail->parentEdge)
	updateSubTreeAfterNNI(A, v, nearEdge->tail->parentEdge, r1, r2, 0.5*dcoeff, DOWN);
      updatePair(A, nearEdge, nearEdge, r1, r2, dcoeff, DOWN);
      A[nearEdge->head->index][v->index] = 
	A[v->index][nearEdge->head->index] = 
	A[v->index][nearEdge->head->index] + 
	0.5*A[nearEdge->head->index][r2->index] -
	0.5*A[nearEdge->head->index][r1->index];
      break;
    case SKEW: /*nearEdge is skew to the center edge of the NNI*/
      if (NULL != nearEdge->head->leftEdge)
	updateSubTreeAfterNNI(A, v, nearEdge->head->leftEdge, r1, r2, 0.5*dcoeff,SKEW);
      if (NULL != nearEdge->head->rightEdge)
	updateSubTreeAfterNNI(A, v, nearEdge->head->rightEdge, r1, r2, 0.5*dcoeff,SKEW);
      updatePair(A, nearEdge, nearEdge, r1, r2, dcoeff, UP);
      A[nearEdge->head->index][v->index] = 
	A[v->index][nearEdge->head->index] = 
	A[v->index][nearEdge->head->index] + 
	0.5*A[nearEdge->head->index][r2->index] -
	0.5*A[nearEdge->head->index][r1->index];
      break;
    }
}


/*swapping across edge whose head is v*/
void bNNIupdateAverages(double **A, node *v, edge *par, edge *skew, 
			edge *swap, edge *fixed)
{  
  node *closer, *further;
  
  A[v->index][v->index] = 0.25*(A[fixed->head->index][par->head->index] + 
				A[fixed->head->index][swap->head->index] + 
				A[skew->head->index][par->head->index] + 
				A[skew->head->index][swap->head->index]);
				
  closer = skew->head;
  further = swap->head;
  updateSubTreeAfterNNI(A, v, fixed, closer, further, 0.25, UP);
  updateSubTreeAfterNNI(A, v, par, further, closer, 0.25, DOWN);
  closer = fixed->head;
  further = par->head;
  updateSubTreeAfterNNI(A, v, skew, closer, further, 0.25, UP);
  updateSubTreeAfterNNI(A, v, swap, further, closer, 0.25, SKEW);
  
}



void bNNItopSwitch(tree *T, edge *e, int direction, double **A)
{
  edge *down, *swap, *fixed;
  node *u, *v;
  if (verbose)
    printf("Performing branch swap across edge %s.\n",e->label);
  down = siblingEdge(e);
  u = e->tail;
  v = e->head;
  if (LEFT == direction)
    {
      swap = e->head->leftEdge;
      fixed = e->head->rightEdge;
      v->leftEdge = down;
    }
  else
    {
      swap = e->head->rightEdge;
      fixed = e->head->leftEdge;
      v->rightEdge = down;
    }
  swap->tail = u;
  down->tail = v;
  if(e->tail->leftEdge == e)
    u->rightEdge = swap;
  else
    u->leftEdge = swap;
  bNNIupdateAverages(A, v, e->tail->parentEdge, down, swap, fixed);
}

double wf5(double D_AD, double D_BC, double D_AC, double D_BD,
	   double D_AB, double D_CD)
{
  double weight;
  weight = 0.25*(D_AC + D_BD + D_AD + D_BC) + 0.5*(D_AB + D_CD);
  return(weight);
}

int bNNIEdgeTest(edge *e, tree *T, double **A, double *weight)
{
  edge *f;
  double D_LR, D_LU, D_LD, D_RD, D_RU, D_DU;
  double w1,w2,w0;
  /*if (verbose)
    printf("Branch swap: testing edge %s.\n",e->label);*/
  if ((leaf(e->tail)) || (leaf(e->head)))
    return(NONE);

  f = siblingEdge(e);

  D_LR = A[e->head->leftEdge->head->index][e->head->rightEdge->head->index];
  D_LU = A[e->head->leftEdge->head->index][e->tail->index];
  D_LD = A[e->head->leftEdge->head->index][f->head->index];
  D_RU = A[e->head->rightEdge->head->index][e->tail->index];
  D_RD = A[e->head->rightEdge->head->index][f->head->index];
  D_DU = A[e->tail->index][f->head->index];

  w0 = wf5(D_RU,D_LD,D_LU,D_RD,D_DU,D_LR);
  w1 = wf5(D_RU,D_LD,D_DU,D_LR,D_LU,D_RD);
  w2 = wf5(D_DU,D_LR,D_LU,D_RD,D_RU,D_LD);
  if (w0 <= w1)
    {
      if (w0 <= w2) /*w0 <= w1,w2*/
	{
	  *weight = 0.0;
	  return(NONE);
	}
      else /*w2 < w0 <= w1 */
	{
	  *weight = w2 - w0;
	  if (verbose)
	    {
	      printf("Possible swap across %s. ",e->label);
	      printf("Weight dropping by %lf.\n",w0 - w2);
	      printf("New weight should be %lf.\n",T->weight + w2 - w0);
	    }
	  return(RIGHT);
	}
    }
  else if (w2 <= w1) /*w2 <= w1 < w0*/
    {
      *weight = w2 - w0;
      if (verbose)
	{
	  printf("Possible swap across %s. ",e->label);
	  printf("Weight dropping by %lf.\n",w0 - w2);
	  printf("New weight should be %lf.\n",T->weight + w2 - w0);
	}
      return(RIGHT);
    }
  else /*w1 < w2, w0*/
    {
      *weight = w1 - w0;
      if (verbose)
	{
	  printf("Possible swap across %s. ",e->label);
	  printf("Weight dropping by %lf.\n",w0 - w1);
	  printf("New weight should be %lf.\n",T->weight + w1 - w0);
	}
      return(LEFT);	
    }
}


  
/*limitedFillTableUp fills all the entries in D associated with
  e->head,f->head and those edges g->head above e->head, working
  recursively and stopping when trigger is reached*/
void limitedFillTableUp(edge *e, edge *f, double **A, edge *trigger)
{
  edge *g,*h;
  g = f->tail->parentEdge;
  if (f != trigger)
    limitedFillTableUp(e,g,A,trigger);
  h = siblingEdge(f);
  A[e->head->index][f->head->index] = 
    A[f->head->index][e->head->index] =  
    0.5*(A[e->head->index][g->head->index] + A[e->head->index][h->head->index]);    
}
  
void WFext(edge *e, double **A) /*works except when e is the one edge
				  inserted to new vertex v by firstInsert*/
{
  edge *f, *g;
  if ((leaf(e->head)) && (leaf(e->tail)))
    e->distance = A[e->head->index][e->head->index];
  else if (leaf(e->head))
    {
      f = e->tail->parentEdge;
      g = siblingEdge(e);
      e->distance = 0.5*(A[e->head->index][g->head->index]
			 + A[e->head->index][f->head->index]
			 - A[g->head->index][f->head->index]);
    }
  else
    {
      f = e->head->leftEdge;
      g = e->head->rightEdge;
      e->distance = 0.5*(A[g->head->index][e->head->index]
			 + A[f->head->index][e->head->index]
			 - A[f->head->index][g->head->index]);
    }
}

void WFint(edge *e, double **A)
{
  int up, down, left, right;
  up = e->tail->index;
  down = (siblingEdge(e))->head->index;
  left = e->head->leftEdge->head->index;
  right = e->head->rightEdge->head->index;
  e->distance = 0.25*(A[up][left] + A[up][right] + A[left][down] + A[right][down]) - 0.5*(A[down][up] + A[left][right]);
} 

  

void assignBalWeights(tree *T, double **A)
{
  edge *e;
  e = depthFirstTraverse(T,NULL);
  while (NULL != e) {
    if ((leaf(e->head)) || (leaf(e->tail)))
      WFext(e,A);
    else
      WFint(e,A);
    e = depthFirstTraverse(T,e);
  }
}      
    
