#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fastme.h"
#include "graph.h"
#include "newick.h"
#include "string.h"

/*functions from newick.c*/
tree *loadNewickTree(FILE *ifile, int numLeaves);
void NewickPrintTree(tree *T, FILE *ofile);
tree *detrifurcate(tree *T);
void partitionSizes(tree *T);

/*functions from bme.c*/
tree *BMEaddSpecies(tree *T,node *v, double **D, double **A); 
void makeBMEAveragesTable(tree *T, double **D, double **A);
void assignBMEWeights(tree *T, double **A);

/*from gme.c*/
tree *GMEaddSpecies(tree *T,node *v, double **D, double **A);
void makeOLSAveragesTable(tree *T, double **D, double **A);
void assignOLSWeights(tree *T, double **A);
void assignAllSizeFields(tree *T);

/*from inputs.c*/
double **loadMatrix(FILE *ifile, int *size, set *S);
void compareSets(tree *T, set *S, FILE *ofile);
void freeMatrix(double **D, int size);

/*from NNI.c*/
void NNI(tree *T, double **avgDistArray, int *count);

/*from bNNI.c*/
void bNNI(tree *T, double **avgDistArray, int *count);

/*from graph.c*/
void freeSet(set *S);
void freeTree(tree *T);

void chooseSettings(int argc, char **argv, int *btype, int *ntype, int *wtype, int *numDataSets, char **filenames)
{
  int counter = 1;
  int i;
  sprintf(filenames[0],"input.d");
  sprintf(filenames[1],"output.t");
  sprintf(filenames[2],"input.t");
  while (counter < argc)
    {
      switch(argv[counter][1])
	{	
	case 'i':
	  counter++;
	  if (NULL != argv[counter])
	    strcpy(filenames[0],argv[counter]);
	  else
	    {
	      fprintf(stderr,"Error: -d flag requires argument.\n");
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 'o':
	  counter++;
	  if (NULL != argv[counter])	    
	    strcpy(filenames[1],argv[counter]);
	  else
	    {
	      fprintf(stderr,"Error: -o flag requires argument.\n");
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 't':
	  counter++;
	  if (NULL != argv[counter])
	    strcpy(filenames[2],argv[counter]);
	  else
	    {
	      fprintf(stderr,"Error: -i flag requires argument.\n");
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  *btype = USER;
	  break;
	case 'w':
	  counter++;
	  switch(argv[counter][0])
	    {
	    case 'b':
	    case 'B':
	      *wtype = BAL;
	      break;
	    case 'g':
	    case 'G':
	    case 'o':
	    case 'O':
	      *wtype = OLS;
	      break;
	    default:
	      fprintf(stderr,"Unknown argument to -w option: please");
	      fprintf(stderr," use (b)alanced or (O)LS\n");
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 'b':
	  counter++;
	  switch(argv[counter][0])
	    {
	    case 'b':
	    case 'B':
	      *btype = BAL;
	      break;
	    case 'g':
	    case 'G':
	    case 'o':
	    case 'O':
	      *btype = OLS;
	      break;
	    default:
	      fprintf(stderr,"Unknown argument to -b option: please");
	      fprintf(stderr," use BME or GME\n");
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 'n':
	  counter++;
	  *numDataSets = i = 0;
	  while (argv[counter][i])
	    *numDataSets = 10* (*numDataSets) + (argv[counter][i++] - '0');
	  counter++;
	  break;
	case 's':
	  counter++;
	  switch(argv[counter][0])
	    {
	    case 'b':
	    case 'B':
	      *ntype = BAL;
	      break;
	    case 'g':
	    case 'G':
	    case 'o':
	    case 'O':
	      *ntype = OLS;
	      break;
	    case 'n':
	    case 'N':
	      *ntype = NONE;
	      break;
	    default:
	      fprintf(stderr,"Unknown argument to -s option: please");
	      fprintf(stderr," use BME, GME, or none\n");
	      exit(EXIT_FAILURE);
	    }
	  counter++;
	  break;
	case 'v':
	  verbose = TRUE;
	  counter++;
	  break;
	case 'h':
	default:
	  fprintf(stderr,"Usage: fastme -binostv\n");
	  fprintf(stderr,"-b specify method for building initial tree: ");
	  fprintf(stderr,"BME or GME(default).\n");
	  fprintf(stderr,"-i filename of distance matrix\n");
	  fprintf(stderr,"-n number of trees/matrices input (default = 1)\n");
	  fprintf(stderr,"-o filename for tree output\n");
	  fprintf(stderr,"-s specify type of tree swapping (NNIs): ");
	  fprintf(stderr,"(b)alanced, (O)LS, or (n)one. (Default is balanced.)\n");
	  fprintf(stderr,"-t (optional) filename of starting tree topology\n");

	  fprintf(stderr,"-v for verbose output\n");
	  fprintf(stderr,"-w (b)alanced or (O)LS weights (if not doing NNIs on input topology) \n");
	  fprintf(stderr,"-help to get this message\n");
	  exit(0);
	}
    }
}

double **initDoubleMatrix(int d)
{
  int i,j;
  double **A;
  A = (double **) malloc(d*sizeof(double *));
  for(i=0;i<d;i++)
    {
      A[i] = (double *) malloc(d*sizeof(double));
      for(j=0;j<d;j++)
	A[i][j] = 0.0;
    }
  return(A);
}


main(int argc, char **argv)
{
  int i;
  int setCounter = 0;
  FILE *ifile1, *ifile2, *ofile;
  double **D, **A;
  set *species, *slooper;
  char **filenames;
  tree *T;
  int numSpecies;
  int numSets = 1;
  int count = 0;
  int nniCount = 0;
  int ntype, btype, wtype;
  ntype = BAL;
  wtype = btype = OLS;
  verbose = FALSE;
  filenames = (char **) malloc (3*sizeof(char *));
  T = NULL;
  for(i=0;i<3;i++)
    filenames[i] = (char *) malloc(MAX_FILE_NAME_LENGTH*sizeof(char));
  chooseSettings(argc,argv,&btype,&ntype,&wtype,&numSets,filenames);
  ifile1 = fopen(filenames[0],"r");
  if (!ifile1)
    {
      fprintf(stderr,"Error opening input file %s\n",filenames[0]);
      exit(EXIT_FAILURE);
    }
  ofile = fopen(filenames[1],"w");
  if (!ofile)
    {
      fprintf(stderr,"Error opening output file %s\n",filenames[1]);
      exit(EXIT_FAILURE);
    }
  if (USER == btype)
    ifile2 = fopen(filenames[2],"r");
  while ( setCounter < numSets ) {
    species = (set *) malloc(sizeof(set));
    species->firstNode = NULL;
    species->secondNode = NULL;
    D = loadMatrix(ifile1,&numSpecies,species);
    A = initDoubleMatrix(2*numSpecies-2);  
    switch(btype)
      {
      case USER:
	if (!ifile2)
	  {
	    fprintf(stderr,"Error opening input file %s\n",filenames[2]);
	    exit(EXIT_FAILURE);
	  }
	T = loadNewickTree(ifile2,numSpecies);
	T = detrifurcate(T);      
	compareSets(T,species,ofile);
	partitionSizes(T);
	break;
      case OLS:
	for(slooper = species; NULL != slooper; slooper = slooper->secondNode)
	  T = GMEaddSpecies(T,slooper->firstNode,D,A);
	break;
      case BAL:
	for(slooper = species; NULL != slooper; slooper = slooper->secondNode)
	  T = BMEaddSpecies(T,slooper->firstNode,D,A);
	break;
      }
    switch(ntype)
      {      
      case OLS:
	if (OLS != btype)
	  assignAllSizeFields(T);
	makeOLSAveragesTable(T,D,A);
	NNI(T,A,&nniCount);
	assignOLSWeights(T,A);
	break;
      case BAL:	
	if (BAL != btype)
	  makeBMEAveragesTable(T,D,A);
	bNNI(T,A,&nniCount);
	assignBMEWeights(T,A);
	break;
      case NONE:
	switch(wtype)
	  {
	  case OLS:
	    if (OLS != btype)
	      assignAllSizeFields(T);
	    makeOLSAveragesTable(T,D,A);
	    assignOLSWeights(T,A);
	    break;
	  case BAL:
	    if (BAL != btype)
	      makeBMEAveragesTable(T,D,A);
	    assignBMEWeights(T,A);
	    break;
	  default:
	    fprintf(stderr,"Error in program: variable 'btype' has illegal ");
	      fprintf(stderr,"value %d.\n",btype);
	      exit(EXIT_FAILURE);
	  }
	break;
      default:
	fprintf(stderr,"Error in program: variable 'ntype' has illegal ");
	fprintf(stderr,"value %d.\n",ntype);
	exit(EXIT_FAILURE);
      }
    NewickPrintTree(T,ofile);
    freeMatrix(D,numSpecies);
    freeMatrix(A,2*numSpecies - 2);
    freeSet(species);
    freeTree(T);
    T = NULL;
    setCounter++;
    if ((verbose) && (ntype))
      printf("Performed %d NNIs on data set %d\n",nniCount,setCounter);
    nniCount = 0;
  }
}
