#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "graph.h"

typedef char boolean;

int *initPerm(int size)
{
  int *p;
  int i;
  p = (int *) malloc(size*sizeof(int));
  for(i = 0;i<size;i++)
    p[i] = i;
  return(p);
}

void permInverse(int *p, int *q, int length)
{
  int i;
  for(i=0;i<length;i++)
    q[p[i]] = i;
}




void insertionSort(int *perm, double *weightArray, int arraySize)
{
  int i,j;
  int temp;
  double a;
  for(j = 1;j<arraySize;j++)
    {
      a = weightArray[perm[j]];
      temp = perm[j];
      i = j-1;
      while((i >= 0) && (weightArray[perm[i]] > a))
	{
	  perm[i+1] = perm[i];
	  i--;
	}
      perm[i+1] = temp;
    }
}  

/*k insertion sort assumes that everything after the first k elements
  has already been sorted*/

void kinsertionSort(int *perm, double *weightArray, int arraySize, int k)
{
  int i,j;
  int temp;
  double a;
  for(j = k-1;j>=0;j--)
    {
      a = weightArray[perm[j]];
      temp = perm[j];
      i = j+1;
      while((i < arraySize) && (weightArray[perm[i]] < a))
	{
	  perm[i-1] = perm[i];
	  i++;
	}
      perm[i-1] = temp;
    }
}  

/*void fsort(int *p, double *v, int size)
{
  if (size < 20)
    insertionSort(p,v,size);
  else
    fquickSort(v,size,p);
}
*/

/*swaps two values of a permutation*/
void swap(int *p, int *q, int i, int j)
{
  int temp;
  temp = p[i];
  p[i] = p[j];
  p[j] = temp;
  q[p[i]] = i;
  q[p[j]] = j;
}

/*The usual Heapify function, tailored for our use with a heap of scores*/
/*will use array p to keep track of indexes*/
/*after scoreHeapify is called, the subtree rooted at i 
  will be a heap*/

/*p goes from heap to array, q goes from array to heap*/

void heapify(int *p, int *q, double *HeapArray, int i, int n)
{
  boolean moreswap = TRUE;

  do {
    int left = 2 * i;
    int right = 2* i + 1;
    int smallest;
    if ((left <= n) &&	(HeapArray[p[left]] < HeapArray[p[i]]))
      smallest = left;
    else
      smallest = i;
    if ((right <= n) && (HeapArray[p[right]] < HeapArray[p[smallest]]))
      smallest = right;
    if (smallest != i){
      swap(p,q,i, smallest);     
      /*push smallest up the heap*/    
      i = smallest;            /*check next level down*/
    }
    else
      moreswap = FALSE;
  } while(moreswap);
}

/*cannot vouch for correctness of next two functions RD 2/7/01 */

/*function builds the heap, with permuation p keeping track of the
order*/
void buildHeap(int *p, int *q, double *HeapArray, int ActualSize)
{
  int i;
  for(i=(ActualSize/2); i>=1; --i)
    heapify(p,q,HeapArray, i, ActualSize);
}

/*function uses heapsort routines to sort a heap of scores*/
void heapSort(double *HeapArray, int ActualSize, int *p, int *q)
{
  int i;
  buildHeap(p,q,HeapArray, ActualSize);
  for(i=ActualSize; i>=2; --i)
    {
      swap(p,q,1, i);
      heapify(p,q,HeapArray, 1, i-1);
    }
}



/*heap is of indices of elements of v, 
  popHeap takes the index at position i and pushes it out of the heap
  (by pushing it to the bottom of the heap, where it is not noticed)*/

void reHeapElement(int *p, int *q, double *v, int length, int i)
{
  int up, here;
  here = i;
  up = i / 2;
  if ((up > 0) && (v[p[here]] < v[p[up]]))
    while ((up > 0) && (v[p[here]] < v[p[up]])) /*we push the new
						  value up the heap*/
      {
	swap(p,q,up,here);
	here = up;
	up = here / 2;
      }
  else
    heapify(p,q,v,i,length);
}

void popHeap(int *p, int *q, double *v, int length, int i)
{
  swap(p,q,i,length); /*puts new value at the last position in the heap*/
  reHeapElement(p,q, v,length-1,i); /*put the swapped guy in the right place*/
}

void pushHeap(int *p, int *q, double *v, int length, int i)
{
  swap(p,q,i,length+1); /*puts new value at the last position in the heap*/
  reHeapElement(p,q, v,length+1,length+1); /*put that guy in the right place*/
}



int makeThreshHeap(int *p, int *q, double *v, int arraySize, double thresh)
{
  int i, heapsize;
  heapsize = 0;
  for(i = 1; i < arraySize;i++)
    if(v[q[i]] < thresh)
      pushHeap(p,q,v,heapsize++,i);
  return(heapsize);
}
